#include <stdio.h>
#include "libgrass.h"
#include "imagery.h"

static void Usage()

{
    printf( "Usage: grassinfo [-db database] [-l location] [-m mapset]\n"
            "                 [-ll] [-lm] [cellname]\n" );
    exit( 1 );
}

int main( int argc, char ** argv )

{
    char	*gisdbase = NULL;
    char	*location = NULL;
    char	*mapset = NULL, *group_name = NULL;
    char	*cellmapset, *user_cellname = NULL, *cellname = NULL;
    char        *cellnames[100];
    char	*p4_def;
    int         cellfile = 0, n, error, data_type, icell;
    int         list_locations = 0, list_mapsets = 0;
    struct Cell_head w;

    /*
     * Handle commandline arguments. 
     */

    for( n = 1; n < argc; n++ )
    {
        if( strcmp(argv[n],"-db") == 0 )
            gisdbase = argv[++n];
        else if( strcmp(argv[n],"-l") == 0 )
            location = argv[++n];
        else if( strcmp(argv[n],"-m") == 0 )
            mapset = argv[++n];
        else if( strcmp(argv[n],"-ll") == 0 )
            list_locations = 1;
        else if( strcmp(argv[n],"-lm") == 0 )
            list_mapsets = 1;
        else if( argv[n][0] == '-' )
            Usage();
        else
            user_cellname = argv[n];
    }

    /*
     * Initialize base on the select (or default) mapset.
     */
    error = G_gisinit_2 (argv[0], gisdbase, location, mapset );

    /* 
     * Handle location list request.
     */
    if( list_locations )
    {
        char  **list;
        int     n;

        list = G_fetch_list_locations( gisdbase );
        
        printf( "Possible location names:\n" );
        for( n = 0; list != NULL && list[n] != NULL; n++ )
            printf( "%s\n", list[n] );
        
        G_free_list( list );
    }

    /* 
     * Handle request for a list of mapsets.
     */
    if( list_mapsets )
    {
        char  **list;
        int     n;

        list = G_fetch_list_mapsets( NULL );
        
        printf( "Possible mapset names:\n" );
        for( n = 0; list != NULL && list[n] != NULL; n++ )
            printf( "%s\n", list[n] );
        
        G_free_list( list );
    }

    if( list_mapsets || list_locations )
        exit( 0 );

    if( error != 0 && user_cellname == NULL )
    {
        fprintf( stderr, 
                 "Unable to initialize database access, no valid .grassrc5\n");
        exit( 1 );
    }

    /*
     * Did the user select a cell file?  If not, list available ones. 
     */
    if( user_cellname == NULL )
    {
        char    **list;
        int     n;

        list = G_fetch_list_element( "cellhd", "" );
        
        printf( "Possible cell names:\n" );
        for( n = 0; list != NULL && list[n] != NULL; n++ )
            printf( "%s\n", list[n] );
        
        G_free_list( list );
        
        list = G_fetch_list_element( "group", "" );
        
        printf( "Possible imagery group names:\n" );
        for( n = 0; list != NULL && list[n] != NULL; n++ )
            printf( "%s\n", list[n] );
        
        G_free_list( list );
        
        exit( 1 );
    }

    /*
     * Translate a group into a list of cells. 
     */

    if( I_check_group( user_cellname, &mapset, &group_name ) )
    {
        struct Ref ref;
        int    iref;

        /* we don't need I_init_group_ref( &ref );, because
         * I_get_group_ref() calls it automatically */
        I_get_group_ref( group_name, &ref );
        printf( "Opening Imagery group `%s'.\n\n", group_name );
        cellnames[0] = NULL;
        for( iref = 0; iref < ref.nfiles; iref++ )
        {
            char    buf[128];

            sprintf( buf, "%s@%s", ref.file[iref].name, ref.file[iref].mapset);
            cellnames[iref] = G_strdup( buf );
            cellnames[iref+1] = NULL;
        }

	I_free_group_ref(&ref);
    }
    else
    {
        cellnames[0] = G_strdup(user_cellname);
        cellnames[1] = NULL;
    }

    /*
     * Check that the requested cell is available.
     */

    for( icell = 0; cellnames[icell] != NULL; icell++ )
    {
        user_cellname = cellnames[icell];

        if( !G_check_cell( user_cellname, &mapset, &cellname )  )
        {
            fprintf(stderr,
                    "Not able to open cellfile for [%s]\n", 
                    user_cellname);
            exit( 1 );
        }

        printf( "\n" );
        printf( "Cell %s\n", cellname );

        G_get_cellhd( cellname, mapset, &w );
        data_type = G_raster_map_type( cellname, mapset );

        if( data_type == CELL_TYPE )
            printf( "data type = integer (CELL)\n" );
        else if( data_type == FCELL_TYPE )
            printf( "data type = float (FCELL)\n" );
        else if( data_type == DCELL_TYPE )
            printf( "data type = double (DCELL)\n" );

        printf( "bytes per pixel = %d\n", w.format+1 );
        printf( "compressed = %d\n", w.compressed );
        printf( "rows, cols = %dx%d\n", w.rows, w.cols );
        printf( "proj = %d\n", w.proj );
        printf( "zone = %d\n", w.zone );
        printf( "ew_res = %f\n", w.ew_res );
        printf( "ns_res = %f\n", w.ns_res );
        printf( "north = %f\n", w.north );
        printf( "south = %f\n", w.south );
        printf( "east = %f\n", w.east );
        printf( "west = %f\n", w.west );

        /* 
         * Try to return a PROJ.4 formatted coordinate system.
         */

        p4_def = G_get_cell_as_proj4( cellname, mapset );
        if( p4_def == NULL )
        {
            printf( "PROJ.4 Definition = <not available>\n" );
        }
        else
        {
            printf( "PROJ.4 Definition = %s\n", p4_def );
            G_free( p4_def );
        }

        /*
         * Actually open the cell.
         */
        G_set_window( &w );

        cellfile = G_open_cell_old_2( user_cellname );
    
        if( cellfile < 0 )
        {
            fprintf(stderr,"Not able to open cellfile for [%s]\n", user_cellname);
            exit( 1 );
        }

        G_close_cell( cellfile );
    }
}
