#include <stdio.h>
#include <stdlib.h>

#include "libgrass.h"
#include "imagery.h"


static unsigned char *ppm_read_file( const char *, int *, int * );
static void write_to_grass( int width, int height, int pixel_step, 
                           unsigned char *src_data, const char *out_cell );

/************************************************************************/
/*                               Usage()                                */
/************************************************************************/
static void Usage()

{
    printf( "Usage: ppm2grass [-db database] [-l location] [-m mapset]\n"
            "                 [-f] [-proj string] [-b north south east west]\n"
            "                 [-image] ppm_filename cellname\n" );
    exit( 1 );
}

/************************************************************************/
/*                                main()                                */
/************************************************************************/

int main( int argc, char ** argv )

{
    char	*gisdbase = NULL;
    char	*location = NULL;
    char	*mapset = NULL, *group_name = NULL;
    char	*cellmapset, *user_cellname = NULL, *cellname = NULL;
    char        *proj = NULL;
    char        *ppm_filename = NULL;
    char        *cellnames[100];
    char	*p4_def;
    int         cellfile = 0, n, error, data_type, icell, force=0;
    int         as_image = 0;
    struct Cell_head cellhd;
    int		ppm_width, ppm_height;
    unsigned char *ppm_data;
    struct Key_Value *proj_info = NULL;
    struct Key_Value *proj_units = NULL;
    double	north=0.0, south=0.0, east=0.0, west=0.0;

    /*
     * Handle commandline arguments. 
     */

    for( n = 1; n < argc; n++ )
    {
        if( strcmp(argv[n],"-db") == 0 )
            gisdbase = argv[++n];
        else if( strcmp(argv[n],"-l") == 0 )
            location = argv[++n];
        else if( strcmp(argv[n],"-m") == 0 )
            mapset = argv[++n];
        else if( strcmp(argv[n],"-image") == 0 )
            as_image = 1;
        else if( strcmp(argv[n],"-proj") == 0 )
            proj = argv[++n];
        else if( strcmp(argv[n],"-f") == 0 )
            force = 1;
        else if( strcmp(argv[n],"-b") == 0 && n+4 < argc )
        {
            north = atof(argv[++n]);
            south = atof(argv[++n]);
            east = atof(argv[++n]);
            west = atof(argv[++n]);
        }
        else if( argv[n][0] == '-' )
            Usage();
        else if( ppm_filename == NULL )
            ppm_filename = argv[n];
        else if( user_cellname == NULL )
            user_cellname = argv[n];
        else
        {
            Usage();
        }
    }

    if( user_cellname == NULL || ppm_filename == NULL )
        Usage();

    /*
     * Read in the entire PPM file into memory.
     */
    ppm_data = ppm_read_file( ppm_filename, &ppm_width, &ppm_height );

    if( north == 0.0 && south == 0.0 )
    {
        north = ppm_height;
        south = 0.0;
        east = ppm_width;
        west = 0.0;
    }

    /*
     * Initialize based on the selected (or default) mapset.
     */
    error = G_gisinit_2 (argv[0], gisdbase, location, mapset );

    if( error != 0 
        && (location == NULL || error != 2) 
        && (mapset == NULL || error != 3) )
    {
        fprintf( stderr, 
                "Unable to initialize database access, no valid .grassrc5?\n");
        exit( 1 );
    }

    /*
     * If available, translate the projection definition into GRASS format.
     */
    if( proj != NULL )
    {
        if( G_proj4_to_kv( proj, &proj_info, &proj_units ) == 0 )
        {
            fprintf( stderr, 
                     "Projection %s not translatable.\n", proj );
            exit( 1 );
        }
    }

    /*
     * If we got error 2, try to create the location.
     */
    if( location != NULL && error == 2 )
    {
        G_set_cellhd_from_projinfo( &cellhd, proj_info );

        cellhd.rows = ppm_height;
        cellhd.cols = ppm_width;
        cellhd.north = north;
        cellhd.south = south;
        cellhd.ns_res = (north - south) / cellhd.rows;
        cellhd.west = west;
        cellhd.east = east;
        cellhd.ew_res = (east - west) / cellhd.cols;

        G__make_location( location, &cellhd, proj_info, proj_units, stdout );

        error = 0;
    }

    /*
     * If we have a projection, does it match our current location? 
     */
    if( !force && proj != NULL )
    {
        struct Key_Value *proj_units_db, *proj_info_db;

        proj_info_db = G_get_projinfo();
        proj_units_db = G_get_projunits();

        if( G_compare_projections( proj_info_db, proj_units_db, 
                                   proj_info, proj_units ) == 0 )
        {
            fprintf( stderr, 
                     "Source and Destination projections or units appear\n"
                     "to not match, aborting conversion.  Use the -f flag\n"
                     "to force conversion ignoring projection, or use -l\n"
                     "to create a new location matching the request.\n" );
            exit( 1 );
        }
    }

    /*
     * Setup the cell header.
     */
    G_get_window( &cellhd );

    cellhd.rows = ppm_height;
    cellhd.cols = ppm_width;
    cellhd.north = north;
    cellhd.south = south;
    cellhd.ns_res = (north - south) / cellhd.rows;
    cellhd.west = west;
    cellhd.east = east;
    cellhd.ew_res = (east - west) / cellhd.cols;

    G_set_window( &cellhd );

    /*
     * If not as_image, then we boil the image down to greyscale.
     */
    if( !as_image )
    {
        int	i, count = ppm_width * ppm_height;

        for( i=0; i < count; i++ )
        {
            ppm_data[i] = 
                (ppm_data[i*3+0] + ppm_data[i*3+1] + ppm_data[i*3+2]) / 3;
        }

        write_to_grass( ppm_width, ppm_height, 1, ppm_data, user_cellname );
    }
    
    /*
     * Otherwise create an image group based on the cell name, and
     * set red, green and blue layers.
     */
    else
    {
        struct Ref ref;
        char	band_name[512];

        I_init_group_ref( &ref );

        sprintf( band_name, "%s_red", user_cellname );
        write_to_grass( ppm_width, ppm_height, 3, ppm_data + 0, band_name );
        I_add_file_to_group_ref( band_name, G_mapset(), &ref );

        sprintf( band_name, "%s_green", user_cellname );
        write_to_grass( ppm_width, ppm_height, 3, ppm_data + 1, band_name );
        I_add_file_to_group_ref( band_name, G_mapset(), &ref );

        sprintf( band_name, "%s_blue", user_cellname );
        write_to_grass( ppm_width, ppm_height, 3, ppm_data + 2, band_name );
        I_add_file_to_group_ref( band_name, G_mapset(), &ref );

        I_put_group_ref( user_cellname, &ref );
        I_free_group_ref( &ref );
    
        /* make this group the current group */
        I_put_group( user_cellname );
    }

    exit( 0 );
}

/************************************************************************/
/*                           write_to_grass()                           */
/************************************************************************/

static void write_to_grass( int ncols, int nrows, int pixel_step, 
                            unsigned char *src_data, const char *out_cell )

{
    RASTER_MAP_TYPE data_type;
    int row, col;
    int cf;
    CELL *cell;

    data_type = CELL_TYPE;
    
/* -------------------------------------------------------------------- */
/*      Create the new raster                                           */
/* -------------------------------------------------------------------- */
    cf = G_open_raster_new((char *)out_cell, data_type);
    if (cf < 0)
    {
        fprintf( stderr, "unable to create raster map %s", out_cell);
        exit( 1 );
    }

/* -------------------------------------------------------------------- */
/*      Write the raster one scanline at a time.                        */
/* -------------------------------------------------------------------- */
    cell = G_allocate_raster_buf(data_type);

    for (row = 1; row <= nrows; row++)
    {
        for( col = 0; col < ncols; col++ )
        {
            cell[col] = *src_data;
            src_data += pixel_step;
        }

        G_put_raster_row (cf, cell, data_type);
    }

/* -------------------------------------------------------------------- */
/*      Cleanup                                                         */
/* -------------------------------------------------------------------- */
    G_close_cell (cf);
}

/************************************************************************/
/*                           ppm_read_file()                            */
/************************************************************************/

static unsigned char *ppm_read_file( const char *ppm_filename, 
                                     int *ppm_width, int *ppm_height )

{
    FILE	*fp;
    char	header[128];
    unsigned char *ppm_data;
    int         data_bytes;

    fp = fopen( ppm_filename, "rb" );

    if( fp == NULL )
    {
        perror( "fopen" );
        return NULL;
    }
    
    if( fgets( header, sizeof(header), fp ) == NULL )
    {
        perror( "fgets" );
        return NULL;
    }

    /* fetch magic number */
    if( header[0] != 'P' || header[1] != '6' )
    {
        fprintf( stderr, "%s is not a binary PPM file.\n", ppm_filename );
        return NULL;
    }

    /* fetch and width and height */
    if( fgets( header, sizeof(header), fp ) == NULL )
    {
        perror( "fgets" );
        return NULL;
    }

    if( sscanf( header, "%d %d", ppm_width, ppm_height ) != 2 )
    {
        fprintf( stderr, "Failed to collected width and height.\n" );
        return NULL;
    }

    /* fetch and ignore maxval */
    if( fgets( header, sizeof(header), fp ) == NULL )
    {
        perror( "fgets" );
        return NULL;
    }

    /* read raw data */
    data_bytes = *ppm_width * *ppm_height * 3;
    ppm_data = (unsigned char *) malloc(data_bytes);

    if( ppm_data == NULL )
    {
        fprintf( stderr, "malloc(%d) failed.\n", data_bytes );
        return NULL;
    }

    if( fread(ppm_data, 1, data_bytes, fp ) != data_bytes )
    {
        fprintf( stderr, "Failed to read expected %d data bytes.\n",
                 data_bytes );
        free( ppm_data );
        return NULL;
    }

    fclose( fp );

    return ppm_data;
}
